-module(case_tracer).

-export([trace/1, pretty_printer/1, newid/1, freshid/0, parse_call/1]).

trace(Call) ->
  	register(ser,spawn(case_tracer,newid,[1])), 
    
	InitialCall = parse_call(Call),
	AExpr = 
		case is_list(InitialCall) of 
		 	true ->
		 		hd(parse_expr(InitialCall++"."));
		 	false ->
		 		InitialCall
		end,
	{call,_,{remote,_,{atom,_,ModName},_},_} = AExpr,
	%io:format("~p\n~p\n",[AExpr,ModName]),
	compile:file(atom_to_list(ModName) ++ ".erl" ,[{parse_transform,case_clause_sender_pt}]),
	%Return = compile:file(atom_to_list(ModName) ++ ".erl" ,[{parse_transform,case_clause_sender_pt}]),
	%io:format("Return: ~p\n",[Return]), 
	register(case_tracer,self()),
	try
		erlang:purge_module(ModName)
	catch 
		_:_ -> ok
	end,
	code:load_abs(atom_to_list(ModName)),
	spawn(fun() -> erl_eval:expr(AExpr,[]),case_tracer!stop end),
	Trace = lists:reverse(receive_loop(0,[])),
	unregister(case_tracer),
	unregister(ser),
	io:format("Trace registered from ~p~n********************************~n",[Call]),
	pretty_printer(Trace).
	%Trace.

receive_loop(Current,Trace) ->
	receive 
		stop ->
			Trace;
		TraceItem = {case_trace, _} ->
			% io:format("entra ~p\n",[TraceItem]),
			receive_loop(Current + 1, [{Current,TraceItem}|Trace]);

		TraceItem = {send_trace, _} ->
			receive_loop(Current + 1, [{Current,TraceItem}|Trace]);

		TraceItem = {{messages, _},_} ->
			receive_loop(Current + 1, [{Current,TraceItem}|Trace]);

		TraceItem = {receive_trace, _} ->
			receive_loop(Current + 1, [{Current,TraceItem}|Trace])
	%****************************************
		 %;
		 %Other ->
		 	%io:format("entra ~p\n",[Other])
	end.


parse_expr(Func) ->
    case erl_scan:string(Func) of
		{ok, Toks, _} ->
		    case erl_parse:parse_exprs(Toks) of
				{ok, _Term} = Res ->
				    Res;
				_Err ->
				    {error, parse_error}
		    end;
		_Err ->
		    {error, parse_error}
    end.
parse_call(ExprString) ->
    case erl_scan:string(ExprString ++ ".") of
		{ok, Toks, _} ->
		    case erl_parse:parse_term(Toks) of
				{ok, Term} ->
				    Term;
				_Err ->
					case erl_parse:parse_exprs(Toks) of
						{ok, [Term]} ->
							Term;
						_Err ->
					    	{error, parse_error}
				    end
		    end;
		_Err ->
		    {error, parse_error}
    end.

newid(N) ->
    receive
    {getid, Pid} -> 
    Pid ! {freshid,N},
    				newid(N+1);
	_ -> newid(N)
    end.
 
freshid() -> 
    ser ! {getid, self()},
    receive 
    	{freshid,Id} ->
        Id
    end.

pretty_printer([]) -> 
	io:format("Trace end.~n");
pretty_printer([{NEvent,{{TEvent = messages,Queue},Pid}}|OtherEvent]) ->
	io:format("Event ~p~n ", [NEvent] ),
	io:format("Process ~p queue before receive ~p~n~n", [Pid,Queue]),
	pretty_printer(OtherEvent);
pretty_printer([{NEvent, {TEvent = receive_trace,{Line, NClauseselected,NMess, Clauseselected,_Allclauses}}}|OtherEvent]) ->
	io:format("Event ~p~n", [NEvent] ),
	io:format("Type ~p line ~p, message number ~p~n", [TEvent, Line, NMess]),
	io:format("Selected clause ~p == ~p~n~n", [NClauseselected,Clauseselected]),
	pretty_printer(OtherEvent);
pretty_printer([{NEvent, {TEvent = case_trace,{Line, NClauseselected, Expression, Clauseselected,_Allclauses}}}|OtherEvent]) ->
	io:format("Event ~p~n", [NEvent] ),
	io:format("Type ~p line ~p related to ~p~n", [TEvent, Line, Expression]),
	io:format("Selected clause ~p == ~p~n~n", [NClauseselected,Clauseselected]),
	pretty_printer(OtherEvent);
pretty_printer([{NEvent, {TEvent,{Pid,Call,{messages, Queue}}}}|OtherEvent]) ->
	io:format("Event ~p~n", [NEvent] ),
	io:format("Type ~p sent by ~p call ~p~n", [TEvent, Pid, Call]),
	io:format("Target process queue ~p ~n~n", [Queue]),
	pretty_printer(OtherEvent);	
pretty_printer([{NEvent, {TEvent,{Line, _}}}|OtherEvent]) ->
	io:format("Event ~p~n", [NEvent] ),
	io:format("Type ~p line ~p ~n", [TEvent, Line]),
	pretty_printer(OtherEvent).
